import re

import numpy as np
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.dates.test_BaseDatetimeTransformer import DatetimeMixinTransformTests
from tubular.dates import DatetimeSinusoidCalculator


@pytest.fixture(scope="module", autouse=True)
def example_transformer():
    return DatetimeSinusoidCalculator("a", "cos", "hour", 24)


class TestInit(
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
):
    """Tests for DatetimeSinusoidCalculator.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeSinusoidCalculator"

    @pytest.mark.parametrize("incorrect_type_method", [2, 2.0, True, {"a": 4}])
    def test_method_type_error(self, incorrect_type_method):
        """Test that an exception is raised if method is not a str or a list."""
        with pytest.raises(
            TypeError,
            match="method must be a string or list but got {}".format(
                type(incorrect_type_method),
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                incorrect_type_method,
                "hour",
                24,
            )

    @pytest.mark.parametrize("incorrect_type_units", [2, 2.0, True, ["help"]])
    def test_units_type_error(self, incorrect_type_units):
        """Test that an exception is raised if units is not a str or a dict."""
        with pytest.raises(
            TypeError,
            match=f"units must be a string or dict but got {type(incorrect_type_units)}",
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                incorrect_type_units,
                24,
            )

    @pytest.mark.parametrize("incorrect_type_period", ["2", True, ["help"]])
    def test_period_type_error(self, incorrect_type_period):
        """Test that an error is raised if period is not an int or a float or a dictionary."""
        with pytest.raises(
            TypeError,
            match="period must be an int, float or dict but got {}".format(
                type(incorrect_type_period),
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                "hour",
                incorrect_type_period,
            )

    @pytest.mark.parametrize(
        "incorrect_dict_types_period",
        [{"str": True}, {2: "str"}, {2: 2}, {"str": ["str"]}],
    )
    def test_period_dict_type_error(self, incorrect_dict_types_period):
        """Test that an error is raised if period dict is not a str:int or str:float kv pair."""
        with pytest.raises(
            TypeError,
            match="period dictionary key value pair must be str:int or str:float but got keys: {} and values: {}".format(
                {type(k) for k in incorrect_dict_types_period},
                {type(v) for v in incorrect_dict_types_period.values()},
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                "hour",
                incorrect_dict_types_period,
            )

    @pytest.mark.parametrize(
        "incorrect_dict_types_units",
        [
            {"str": True},
            {2: "str"},
            {"str": 2},
            {2: 2},
            {"str": True},
            {"str": ["str"]},
        ],
    )
    def test_units_dict_type_error(self, incorrect_dict_types_units):
        """Test that an error is raised if units dict is not a str:str kv pair."""
        with pytest.raises(
            TypeError,
            match="units dictionary key value pair must be strings but got keys: {} and values: {}".format(
                {type(k) for k in incorrect_dict_types_units},
                {type(v) for v in incorrect_dict_types_units.values()},
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                incorrect_dict_types_units,
                24,
            )

    @pytest.mark.parametrize("incorrect_dict_units", [{"str": "tweet"}])
    def test_units_dict_value_error(self, incorrect_dict_units):
        """Test that an error is raised if units dict value is not from the valid units list."""
        with pytest.raises(
            ValueError,
            match="units dictionary values must be one of 'year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond' but got {}".format(
                set(incorrect_dict_units.values()),
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                incorrect_dict_units,
                24,
            )

    @pytest.mark.parametrize(
        "incorrect_dict_columns_period",
        [{"ham": 24}, {"str": 34.0}],
    )
    def test_period_dict_col_error(self, incorrect_dict_columns_period):
        """Test that an error is raised if period dict keys are not equal to columns."""
        with pytest.raises(
            ValueError,
            match="period dictionary keys must be the same as columns but got {}".format(
                set(incorrect_dict_columns_period.keys()),
            ),
        ):
            DatetimeSinusoidCalculator(
                ["vegan_sausages", "carrots", "peas"],
                "cos",
                "hour",
                incorrect_dict_columns_period,
            )

    @pytest.mark.parametrize(
        "incorrect_dict_columns_unit",
        [{"sausage_roll": "hour"}],
    )
    def test_unit_dict_col_error(self, incorrect_dict_columns_unit):
        """Test that an error is raised if unit dict keys is not equal to columns."""
        with pytest.raises(
            ValueError,
            match="unit dictionary keys must be the same as columns but got {}".format(
                set(incorrect_dict_columns_unit.keys()),
            ),
        ):
            DatetimeSinusoidCalculator(
                ["vegan_sausages", "carrots", "peas"],
                "cos",
                incorrect_dict_columns_unit,
                6,
            )

    def test_valid_method_value_error(self):
        """Test that a value error is raised if method is not sin, cos or a list containing both."""
        method = "tan"

        with pytest.raises(
            ValueError,
            match=f'Invalid method {method} supplied, should be "sin", "cos" or a list containing both',
        ):
            DatetimeSinusoidCalculator(
                "a",
                method,
                "year",
                24,
            )

    def test_valid_units_value_error(self):
        """Test that a value error is raised if the unit supplied is not in the valid units list."""
        units = "five"
        valid_unit_list = [
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "microsecond",
        ]

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Invalid units {units} supplied, should be in {valid_unit_list}",
            ),
        ):
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                units,
                24,
            )


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeSinusoidCalculator"


class TestTransform(GenericTransformTests, DatetimeMixinTransformTests):
    """Tests for BaseTwoColumnDateTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeSinusoidCalculator"

    @pytest.mark.parametrize(
        "transformer",
        [
            DatetimeSinusoidCalculator(
                "a",
                "cos",
                "month",
                12,
            ),
            DatetimeSinusoidCalculator(
                [
                    "a",
                    "b",
                ],
                "cos",
                "month",
                12,
            ),
        ],
    )
    def test_expected_output_single_method(self, transformer):
        expected = d.create_datediff_test_df()
        for column in transformer.columns:
            column_in_desired_unit = expected[column].dt.month
            cos_argument = column_in_desired_unit * (2.0 * np.pi / 12)
            new_col_name = "cos_12_month_" + column
            expected[new_col_name] = cos_argument.apply(np.cos)

        x = transformer
        actual = x.transform(d.create_datediff_test_df())
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="DatetimeSinusoidCalculator transformer does not produce the expected output",
        )

    def test_expected_output_both_methods(self):
        expected = d.create_datediff_test_df()
        columns = ["a", "b"]
        transformer = DatetimeSinusoidCalculator(
            method=["sin", "cos"],
            units="month",
            period=12,
            columns=columns,
        )

        for column in transformer.columns:
            column_in_desired_unit = expected[column].dt.month
            method_ready_column = column_in_desired_unit * (2.0 * np.pi / 12)
            new_cos_col_name = "cos_12_month_" + column
            new_sin_col_name = "sin_12_month_" + column
            expected[new_sin_col_name] = method_ready_column.apply(np.sin)
            expected[new_cos_col_name] = method_ready_column.apply(np.cos)

        df = d.create_datediff_test_df()
        actual = transformer.transform(df)
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="DatetimeSinusoidCalculator transformer does not produce the expected output",
        )

    def test_expected_output_dict_units(self):
        expected = d.create_datediff_test_df()
        columns = ["a", "b"]
        transformer = DatetimeSinusoidCalculator(
            columns=columns,
            method=["sin"],
            units={"a": "month", "b": "day"},
            period=12,
        )

        a_in_desired_unit = expected["a"].dt.month
        b_in_desired_unit = expected["b"].dt.day
        a_method_ready_column = a_in_desired_unit * (2.0 * np.pi / 12)
        b_method_ready_column = b_in_desired_unit * (2.0 * np.pi / 12)
        a_col_name = "sin_12_month_a"
        b_col_name = "sin_12_day_b"
        expected[a_col_name] = a_method_ready_column.apply(np.sin)
        expected[b_col_name] = b_method_ready_column.apply(np.sin)
        df = d.create_datediff_test_df()
        actual = transformer.transform(df)
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="DatetimeSinusoidCalculator transformer does not produce the expected output",
        )

    def test_expected_output_dict_period(self):
        expected = d.create_datediff_test_df()

        transformer = DatetimeSinusoidCalculator(
            columns=[
                "a",
                "b",
            ],
            method=["sin"],
            units="month",
            period={"a": 12, "b": 24},
        )

        a_in_desired_unit = expected["a"].dt.month
        b_in_desired_unit = expected["b"].dt.month
        a_method_ready_column = a_in_desired_unit * (2.0 * np.pi / 12)
        b_method_ready_column = b_in_desired_unit * (2.0 * np.pi / 24)
        a_col_name = "sin_12_month_a"
        b_col_name = "sin_24_month_b"
        expected[a_col_name] = a_method_ready_column.apply(np.sin)
        expected[b_col_name] = b_method_ready_column.apply(np.sin)

        actual = transformer.transform(d.create_datediff_test_df())
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="DatetimeSinusoidCalculator transformer does not produce the expected output",
        )

    def test_expected_output_dict_both(self):
        expected = d.create_datediff_test_df()
        columns = ["a", "b"]
        transformer = DatetimeSinusoidCalculator(
            columns=columns,
            method=["sin"],
            units={"a": "month", "b": "day"},
            period={"a": 12, "b": 24},
        )

        a_in_desired_unit = expected["a"].dt.month
        b_in_desired_unit = expected["b"].dt.day
        a_method_ready_column = a_in_desired_unit * (2.0 * np.pi / 12)
        b_method_ready_column = b_in_desired_unit * (2.0 * np.pi / 24)
        a_col_name = "sin_12_month_a"
        b_col_name = "sin_24_day_b"
        expected[a_col_name] = a_method_ready_column.apply(np.sin)
        expected[b_col_name] = b_method_ready_column.apply(np.sin)
        df = d.create_datediff_test_df()

        actual = transformer.transform(df)
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="DatetimeSinusoidCalculator transformer does not produce the expected output",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeSinusoidCalculator"



import pandas as pd
from sklearn.datasets import fetch_california_housing


def create_standard_pandas_dataset() -> pd.DataFrame:
    "Function to create the california housing dataset and alter it so that it can be transformed on by every transformer in fubular"

    # Load dataset
    cali = fetch_california_housing()
    cali_df = pd.DataFrame(cali["data"], columns=cali["feature_names"])

    # Sample to introduce missing vals
    cali_df["AveOccup"] = cali_df["AveOccup"].sample(frac=0.99, random_state=1)
    cali_df["HouseAge"] = cali_df["HouseAge"].sample(frac=0.95, random_state=2)
    cali_df["Population"] = cali_df["Population"].sample(frac=0.995, random_state=3)

    # Create a mock categorical feature same length as cali dataset
    cat_list = ["a", "b", "c", "d", "e", "f", "g", "d", "h", "a"] * 2064
    # Introduce rare level
    cat_list[20] = "z"
    # Fewer categorical leveles for OHE
    cat_list_ohe = ["a", "b"] * 2064 * 5

    # Creating copies of levels so that later transformers still have e.g. missing values to impute
    for i in ["1", "2", "3", "4"]:
        cali_df["categorical_" + i] = cat_list

    cali_df["categorical_ohe"] = cat_list_ohe

    for i in range(1, 8):
        str_i = str(i)
        cali_df[
            ["HouseAge_" + str_i, "AveOccup_" + str_i, "Population_" + str_i]
        ] = cali_df[["HouseAge", "AveOccup", "Population"]].copy()

    return cali_df



import narwhals as nw
import polars as pl
import pytest

from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.utils import assert_frame_equal_dispatch


class TestInit(ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseTransformer"


class TestTransform(GenericTransformTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseTransformer"

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_X_returned(self, minimal_dataframe_lookup, initialized_transformers):
        """Test that X is returned from transform."""
        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # if transformer is not polars compatible, skip polars test
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        df = nw.from_native(df)
        expected = df.clone()

        df = nw.to_native(df)
        expected = nw.to_native(expected)

        df_transformed = x.transform(X=df)

        assert_frame_equal_dispatch(expected, df_transformed)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseTransformer"



import pandas as pd


def one_way_summary_plot(df: pd.DataFrame, column: str, response: str = "y") -> None:
    """Function to produce a rough one-way summary plot of a specific column.

    Specifically plot averge response (right y axis) and number of records (left
    y axis) by the selected column (x axis).

    """
    agg = df.groupby(column).agg({column: ["count"], response: ["mean"]})

    ax = agg.plot.bar(y=(column, "count"), ylabel="count", figsize=(8, 5))

    agg.plot(
        y=(response, "mean"),
        style=":",
        marker=".",
        c="k",
        ax=ax,
        use_index=False,
        secondary_y=True,
        mark_right=False,
        rot=90,
        title=f"one-way summary of {column}",
    )

    ax.right_ax.set_ylabel("mean y")



import re

import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tubular.base import DataFrameMethodTransformer


class DataFrameMethodTransformerInitTests(ColumnStrListInitTests):
    """Inheritable tests for DataFrameMethodTransformer.init()."""

    @pytest.mark.parametrize("not_dictionary", ["a", [1, 2], True, 1.5])
    def test_exception_raised_pd_method_kwargs_not_dict(self, not_dictionary):
        """Test an exception is raised if pd_method_kwargs not a dict"""

        with pytest.raises(
            TypeError,
            match=f"DataFrameMethodTransformer: pd_method_kwargs should be a dict but got type {type(not_dictionary)}",
        ):
            DataFrameMethodTransformer(
                new_column_names="a",
                pd_method_name="b",
                columns=["b", "c"],
                pd_method_kwargs=not_dictionary,
            )

    @pytest.mark.parametrize("not_string", [1, True, 1.5])
    def test_exception_raised_pd_method_kwargs_key_not_string(self, not_string):
        """Test an exception is raised if a pd_method_kwarg key is not a string"""

        pd_method_kwargs = {
            "other": 2,
            not_string: 1,
        }

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"DataFrameMethodTransformer: unexpected type ({type(not_string)}) for pd_method_kwargs key in position {1}, must be str",
            ),
        ):
            DataFrameMethodTransformer(
                new_column_names="a",
                pd_method_name="b",
                columns=["b", "c"],
                pd_method_kwargs=pd_method_kwargs,
            )

    @pytest.mark.parametrize("not_string", [{"a": 1}, [1, 2], 1, True, 1.5])
    def test_exception_raised_pd_method_name_not_string(self, not_string):
        """Test an exception is raised if pd_method_name is not a string"""

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"DataFrameMethodTransformer: unexpected type ({type(not_string)}) for pd_method_name, expecting str",
            ),
        ):
            DataFrameMethodTransformer(
                new_column_names="a",
                pd_method_name=not_string,
                columns=["b", "c"],
            )

    def test_exception_raised_non_pandas_method_passed(self):
        """Test an exception is raised if a non pd.DataFrame method is passed for pd_method_name."""
        with pytest.raises(
            AttributeError,
            match="""DataFrameMethodTransformer: error accessing "b" method on pd.DataFrame object - pd_method_name should be a pd.DataFrame method""",
        ):
            DataFrameMethodTransformer(
                new_column_names="a",
                pd_method_name="b",
                columns=["b", "c"],
            )


class TestInit(DropOriginalInitMixinTests, DataFrameMethodTransformerInitTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DataFrameMethodTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DataFrameMethodTransformer"


class TestTransform(DropOriginalTransformMixinTests, GenericTransformTests):
    """Tests for DataFrameMethodTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DataFrameMethodTransformer"

    def expected_df_1():
        """Expected output of test_expected_output_single_columns_assignment."""
        return pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 5, 6, np.nan],
                "b": [1, 2, 3, np.nan, 7, 8, 9],
                "c": [np.nan, 1, 2, 3, -4, -5, -6],
                "d": [1.0, 3.0, 5.0, 3.0, 3.0, 3.0, 3.0],
            },
        )

    def expected_df_2():
        """Expected output of test_expected_output_multi_columns_assignment."""
        return pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 5, 6, np.nan],
                "b": [1, 2, 3, np.nan, 7, 8, 9],
                "c": [np.nan, 1, 2, 3, -4, -5, -6],
                "d": [0.5, 1.0, 1.5, np.nan, 3.5, 4.0, 4.5],
                "e": [np.nan, 0.5, 1.0, 1.5, -2.0, -2.5, -3.0],
            },
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_1()),
    )
    def test_expected_output_single_columns_assignment(self, df, expected):
        """Test a single column output from transform gives expected results."""
        x = DataFrameMethodTransformer(
            new_column_names="d",
            pd_method_name="sum",
            columns=["b", "c"],
            pd_method_kwargs={"axis": 1},
        )

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="DataFrameMethodTransformer sum columns b and c",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_2()),
    )
    def test_expected_output_multi_columns_assignment(self, df, expected):
        """Test a multiple column output from transform gives expected results."""
        x = DataFrameMethodTransformer(
            new_column_names=["d", "e"],
            pd_method_name="div",
            columns=["b", "c"],
            pd_method_kwargs={"other": 2},
        )

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="DataFrameMethodTransformer divide by 2 columns b and c",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseTransformer"



from __future__ import annotations

import pandas as pd  # noqa: TCH002

from tubular.base import BaseTransformer
from tubular.mixins import DropOriginalMixin, NewColumnNameMixin, TwoColumnMixin


class EqualityChecker(
    DropOriginalMixin,
    NewColumnNameMixin,
    TwoColumnMixin,
    BaseTransformer,
):
    """Transformer to check if two columns are equal.

    Parameters
    ----------
    columns: list
        List containing names of the two columns to check.

    new_column_name: string
        string containing the name of the new column.

    drop_original: boolean = False
        boolean representing dropping the input columns from X after checks.

    **kwargs:
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    def __init__(
        self,
        columns: list,
        new_column_name: str,
        drop_original: bool = False,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        self.check_two_columns(columns)
        self.set_drop_original_column(drop_original)
        self.check_and_set_new_column_name(new_column_name)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Create a column which is populated by the boolean
        matching between two columns iterated over rows.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with additional boolean column.

        """
        X = super().transform(X)

        X[self.new_column_name] = X[self.columns[0]] == X[self.columns[1]]

        # Drop original columns if self.drop_original is True
        return DropOriginalMixin.drop_original_column(
            self,
            X,
            self.drop_original,
            self.columns,
        )



"""This module contains functions that create simple datasets that are used in the tests."""

import datetime

import narwhals as nw
import numpy as np
import pandas as pd
import polars as pl

import tests.utils as u


def create_numeric_df_1(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [34.48, 21.71, 32.83, 1.08, 32.93, 4.74, 2.76, 75.7, 14.08, 61.31],
        "b": [12.03, 20.32, 24.12, 24.18, 68.99, 0.0, 0.0, 59.46, 11.02, 60.68],
        "c": [17.06, 12.25, 19.15, 29.73, 1.98, 8.23, 15.22, 20.59, 3.82, 39.73],
        "d": [25.94, 70.22, 72.94, 64.55, 0.41, 13.62, 30.22, 4.6, 67.13, 10.38],
        "e": [94.3, 4.18, 51.7, 16.63, 2.6, 16.57, 3.51, 30.79, 66.19, 25.44],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_numeric_df_2(library="pandas"):
    """Example with numeric dataframe that includes missings."""

    df_dict = {
        "a": [2, 3, 2, 1, 4, 1],
        "b": [None, None, 1, 3, 3, 4],
        "c": [1, 1, 2, 3, 3, None],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_object_df(library="pandas"):
    """Example with object columns - c is numeric target"""

    df_dict = {
        "a": [1, 2, 3, 4, 5],
        "b": ["f", "g", "h", "i", "j"],
        "c": ["a", "b", "c", "d", "e"],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_df_1(library="pandas"):
    """Create simple DataFrame with the following...

    6 rows
    2 columns;
    - a integer 1:6
    - b object a:f
    no nulls
    """

    df_dict = {"a": [1, 2, 3, 4, 5, 6], "b": ["a", "b", "c", "d", "e", "f"]}

    return u.dataframe_init_dispatch(df_dict, library)


def create_df_2(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3, 4, 5, 6, None],
        "b": ["a", "b", "c", "d", "e", "f", None],
        "c": ["a", "b", "c", "d", "e", "f", None],
    }

    df = u.dataframe_init_dispatch(df_dict, library)
    if library == "pandas":
        df["c"] = df["c"].astype("category")
    elif library == "polars":
        df = df.with_columns(df["c"].cast(pl.Categorical))
    return df


def create_df_3(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3, 4, 5, 6, None],
        "b": [1, 2, 3, None, 7, 8, 9],
        "c": [None, 1, 2, 3, -4, -5, -6],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_df_4(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3, 4, 5, 6, 7, None],
        "b": ["a", "b", "c", "d", "e", "f", "g", None],
        "c": ["a", "b", "c", "d", "e", "f", "g", None],
    }

    df = u.dataframe_init_dispatch(df_dict, library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

    return df.to_native()


def create_df_5(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, None],
        "b": ["a", "a", "a", "d", "e", "f", "g", None, None, None],
        "c": ["a", "a", "c", "c", "e", "e", "f", "g", "h", None],
    }

    df = u.dataframe_init_dispatch(df_dict, library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

    return df.to_native()


def create_df_6(library="pandas"):
    """Nulls in different positions to check summing weights by col with nulls."""

    df_dict = {
        "a": [2, 2, 2, 2, 0, 2, 2, 2, 3, 3],
        "b": ["a", "a", "a", "d", "e", "f", "g", None, None, None],
        "c": ["a", "b", "c", "d", "f", "f", "f", "g", "g", None],
    }

    df = u.dataframe_init_dispatch(df_dict, library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

    return df.to_native()


def create_df_7(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [4, 2, 2, 1, 3],
        "b": ["x", "z", "y", "x", "x"],
        "c": ["c", "a", "a", "c", "b"],
    }

    df = u.dataframe_init_dispatch(df_dict, library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

    return df.to_native()


def create_df_8(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 5, 2, 3, 3],
        "b": ["w", "w", "z", "y", "x"],
        "c": ["a", "a", "c", "b", "a"],
    }

    df = u.dataframe_init_dispatch(df_dict, library)

    df = nw.from_native(df)

    df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

    return df.to_native()


def create_df_9(library="pandas"):
    """Create simple DataFrame to use in other tests."""
    df_dict = {
        "a": [1, 2, None, 4, None, 6],
        "b": [None, 5, 4, 3, 2, 1],
        "c": [3, 2, 1, 4, 5, 6],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_df_10(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, None],
        "b": ["a", "a", "a", "d", "e", "f", "g", None, None, None],
        "c": [1, 1, 3, 4, 5, 6, 5, 8, 9, 50],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_df_11(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1, 2, 3],
        "b": [4, 5, 6],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_bool_and_float_df(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [1.0, 2.0, None],
        "b": [True, False, None],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_df_with_none_and_nan_cols(library="pandas"):
    """Create simple DataFrame to use in other tests."""

    df_dict = {
        "a": [np.nan, np.nan, np.nan],
        "b": [None, None, None],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_weighted_imputers_test_df(library="pandas"):
    """Create DataFrame to use imputer tests that correct values are imputed for weighted dataframes.

    weight column contains the weights between 0 and 1
    """

    df_dict = {
        "a": [1.0, 1.0, 1.0, 3.0, 5.0, 5.0],
        "b": ["a", "a", "a", "d", "e", "e"],
        "c": ["a", "a", None, None, None, "f"],
        "d": [1.0, 5.0, 3.0, None, None, 1.0],
        "response": [0, 1, 0, 1, 1, 1],
        "weights_column": [0.1, 0.1, 0.8, 0.5, 0.9, 0.8],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_date_test_df(library="pandas"):
    """Create DataFrame for DateDiffLeapYearTransformer tests."""

    df_dict = {
        "a": [
            datetime.date(1993, 9, 27),
            datetime.date(2000, 3, 19),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 12, 10),
            datetime.date(1985, 7, 23),
        ],
        "b": [
            datetime.date(2020, 5, 1),
            datetime.date(2019, 12, 25),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 9, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 7, 23),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_datediff_test_df(library="pandas"):
    """Create DataFrame for DateDifferenceTransformer tests."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2000,
                3,
                19,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                10,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                12,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            datetime.datetime(2020, 5, 1, 12, 59, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(2018, 9, 10, 9, 59, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(
                2015,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


# Create between dates df staying here as used in conftest also
def create_is_between_dates_df_1(library="pandas"):
    """Create df to use in IsBetweenDates tests. Contains 3 columns of 2 datetime values."""

    df_dict = {
        "a": pd.date_range(start="1/1/2016", end="27/02/2017", periods=2),
        "b": pd.date_range(start="1/2/2016", end="27/09/2017", periods=2),
        "c": pd.date_range(start="1/3/2016", end="27/04/2017", periods=2),
        "d": pd.date_range(start="1/3/2015", end="27/04/2018", periods=2),
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_is_between_dates_df_2(library="pandas"):
    """Create df to use in IsBetweenDates tests. Contains 3 columns of 5 datatime values, covers edge cases."""

    df_dict = {
        "a": [
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
        ],
        "b": [
            datetime.datetime(1990, 1, 20, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 2, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 6, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 2, tzinfo=datetime.timezone.utc),
        ],
        "c": [
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_is_between_dates_df_3(library="pandas"):
    """Create df to use in IsBetweenDates tests. Contains 3 columns of 5 datatime values, covers edge cases, has mixed date data types."""

    df_dict = {
        "a_date": [
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
        ],
        "b_date": [
            datetime.datetime(1990, 1, 20, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 2, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 6, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 2, tzinfo=datetime.timezone.utc),
        ],
        "c_date": [
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
        ],
        "a_datetime": [
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
        ],
        "b_datetime": [
            datetime.datetime(1990, 1, 20, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 2, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 2, 6, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 2, tzinfo=datetime.timezone.utc),
        ],
        "c_datetime": [
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(1990, 3, 1, tzinfo=datetime.timezone.utc),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_date_diff_different_dtypes(library="pandas"):
    """Dataframe with different datetime formats"""

    df_dict = {
        "date_col_1": [
            datetime.date(1993, 9, 27),
            datetime.date(2000, 3, 19),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 12, 10),
            datetime.date(
                1985,
                7,
                23,
            ),
        ],
        "date_col_2": [
            datetime.date(2020, 5, 1),
            datetime.date(2019, 12, 25),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 9, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 7, 23),
        ],
        "datetime_col_1": [
            datetime.datetime(1993, 9, 27, tzinfo=datetime.timezone.utc),
            datetime.datetime(2000, 3, 19, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 12, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(
                1985,
                7,
                23,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "datetime_col_2": [
            datetime.datetime(2020, 5, 1, tzinfo=datetime.timezone.utc),
            datetime.datetime(2019, 12, 25, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 9, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 7, 23, tzinfo=datetime.timezone.utc),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def create_date_diff_different_dtypes_and_nans(library="pandas"):
    """Dataframe with different datetime formats with nans in the data"""

    df_dict = {
        "date_col_1": [
            None,
            datetime.date(2000, 3, 19),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 10, 10),
            datetime.date(2018, 12, 10),
            datetime.date(
                1985,
                7,
                23,
            ),
        ],
        "date_col_2": [
            datetime.date(2020, 5, 1),
            datetime.date(2019, 12, 25),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 9, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 7, 23),
        ],
        "datetime_col_1": [
            datetime.datetime(1993, 9, 27, tzinfo=datetime.timezone.utc),
            datetime.datetime(2000, 3, 19, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 10, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 12, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(
                1985,
                7,
                23,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "datetime_col_2": [
            None,
            datetime.datetime(2019, 12, 25, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2018, 9, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 11, 10, tzinfo=datetime.timezone.utc),
            datetime.datetime(2015, 7, 23, tzinfo=datetime.timezone.utc),
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)


def expected_date_diff_df_2(library="pandas"):
    """Expected output for test_expected_output_drop_cols_true."""

    df_dict = {
        "c": [
            None,
            19,
            0,
            0,
            0,
            -2,
            -3,
            30,
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library=library)



import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    SeparatorInitMixintests,
)
from tubular.strings import StringConcatenator


class TestInit(
    SeparatorInitMixintests,
    ColumnStrListInitTests,
    NewColumnNameInitMixintests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "StringConcatenator"


class TestTransform(GenericTransformTests):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "StringConcatenator"

    def test_correct_df_returned_1(self):
        """Test that correct df is returned after transformation."""
        df = d.create_df_1()
        expected_df = df.copy()

        x = StringConcatenator(
            columns=["a", "b"],
            new_column_name="merged_values",
        )

        df_transformed = x.transform(df)
        expected_df["merged_values"] = ["1 a", "2 b", "3 c", "4 d", "5 e", "6 f"]

        ta.equality.assert_frame_equal_msg(
            df_transformed,
            expected_df,
            "Incorrect dataframe returned after StringConcatenator transform",
        )

    def test_correct_df_returned_2(self):
        """Test that correct df is returned after transformation."""
        df = d.create_df_1()
        expected_df = df.copy()

        x = StringConcatenator(
            columns=["a", "b"],
            new_column_name="merged_values",
            separator=":",
        )

        df_transformed = x.transform(df)
        expected_df["merged_values"] = ["1:a", "2:b", "3:c", "4:d", "5:e", "6:f"]

        ta.equality.assert_frame_equal_msg(
            df_transformed,
            expected_df,
            "Incorrect dataframe returned after StringConcatenator transform",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "StringConcatenator"



import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import OtherBaseBehaviourTests
from tests.mapping.test_BaseCrossColumnNumericTransformer import (
    BaseCrossColumnNumericTransformerInitTests,
    BaseCrossColumnNumericTransformerTransformTests,
)


class TestInit(BaseCrossColumnNumericTransformerInitTests):
    """Tests for CrossColumnAddTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnAddTransformer"


class TestTransform(BaseCrossColumnNumericTransformerTransformTests):
    """Tests for CrossColumnAddTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnAddTransformer"

    def expected_df_1():
        """Expected output from test_expected_output."""
        return pd.DataFrame(
            {"a": [2.1, 3.2, 4.3, 5.4, 6.5, 7.6], "b": ["a", "b", "c", "d", "e", "f"]},
        )

    def expected_df_3():
        """Expected output from test_multiple_mappings_expected_output."""
        df = pd.DataFrame(
            {
                "a": [4.1, 5.1, 4.1, 4, 8, 10.2, 7, 8, 9, np.nan],
                "b": ["a", "a", "a", "d", "e", "f", "g", np.nan, np.nan, np.nan],
                "c": ["a", "a", "c", "c", "e", "e", "f", "g", "h", np.nan],
            },
        )

        df["c"] = df["c"].astype("category")

        return df

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), expected_df_1()),
    )
    def test_expected_output(
        self,
        df,
        expected,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that transform is giving the expected output."""
        mapping = {"b": {"a": 1.1, "b": 1.2, "c": 1.3, "d": 1.4, "e": 1.5, "f": 1.6}}

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping
        args["adjust_column"] = "a"

        x = uninitialized_transformers[self.transformer_name](**args)

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="expected output from cross column add transformer",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_5(), expected_df_3()),
    )
    def test_multiple_mappings_expected_output(
        self,
        df,
        expected,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that mappings by multiple columns are both applied in transform."""
        mapping = {"b": {"a": 1.1, "f": 1.2}, "c": {"a": 2, "e": 3}}

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping
        args["adjust_column"] = "a"

        x = uninitialized_transformers[self.transformer_name](**args)

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="expected output from cross column add transformer",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnAddTransformer"



import numpy as np
import pandas as pd
import pytest
import test_aide as ta

from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tubular.misc import ColumnDtypeSetter


class TestInit(ColumnStrListInitTests):
    """Generic tests for ColumnDtypeSetter.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ColumnDtypeSetter"

    @pytest.mark.parametrize(
        "invalid_dtype",
        ["STRING", "misc_invalid", "np.int", 0],
    )
    def test_invalid_dtype_error(self, invalid_dtype):
        msg = f"ColumnDtypeSetter: data type '{invalid_dtype}' not understood as a valid dtype"
        with pytest.raises(TypeError, match=msg):
            ColumnDtypeSetter(columns=["a"], dtype=invalid_dtype)


class TestFit(GenericFitTests):
    """Generic tests for ColumnDtypeSetter.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ColumnDtypeSetter"


class TestTransform(GenericTransformTests):
    """Tests for ColumnDtypeSetter.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ColumnDtypeSetter"

    def base_df():
        """Input dataframe from test_expected_output."""
        return pd.DataFrame(
            {
                "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, np.nan],
                "b": [1.0, 2.0, 3.0, np.nan, 7.0, 8.0, 9.0],
                "c": [1.0, 1.0, 2.0, 3.0, -4.0, -5.0, -6.0],
                "d": [1, 1, 2, 3, -4, -5, -6],
            },
        )

    def expected_df():
        """Expected output from test_expected_output."""
        return pd.DataFrame(
            {
                "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, np.nan],
                "b": [1.0, 2.0, 3.0, np.nan, 7.0, 8.0, 9.0],
                "c": [1.0, 1.0, 2.0, 3.0, -4.0, -5.0, -6.0],
                "d": [1.0, 1.0, 2.0, 3.0, -4.0, -5.0, -6.0],
            },
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.row_by_row_params(base_df(), expected_df())
        + ta.pandas.index_preserved_params(base_df(), expected_df()),
    )
    @pytest.mark.parametrize("dtype", [float, "float"])
    def test_expected_output(self, df, expected, dtype):
        """Test values are correctly set to float dtype."""
        df["a"] = df["a"].astype(str)
        df["b"] = df["b"].astype(float)
        df["c"] = df["c"].astype(int)
        df["d"] = df["d"].astype(str)

        x = ColumnDtypeSetter(columns=["a", "b", "c", "d"], dtype=dtype)

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="Check values correctly converted to float",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for ColumnDtypeSetter behaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ColumnDtypeSetter"



import narwhals as nw
import pytest

import tests.test_data as d
from tests.mapping.test_BaseMappingTransformer import (
    BaseMappingTransformerInitTests,
    BaseMappingTransformerTransformTests,
    GenericFitTests,
    OtherBaseBehaviourTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.mapping import MappingTransformer


def expected_df_1(library="pandas"):
    """Expected output for test_expected_output."""

    df_dict = {"a": ["a", "b", "c", "d", "e", "f"], "b": [1, 2, 3, 4, 5, 6]}

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)

    df = df.with_columns(nw.col("b").cast(nw.Int8))

    return df.to_native()


def expected_df_2(library="pandas"):
    """Expected output for test_non_specified_values_unchanged."""

    df_dict = {"a": [5, 6, 7, 4, 5, 6], "b": ["z", "y", "x", "d", "e", "f"]}

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)

    df = df.with_columns(nw.col("a").cast(nw.Int8))

    return df.to_native()


class TestInit(BaseMappingTransformerInitTests):
    """Tests for MappingTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MappingTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MappingTransformer"


class TestTransform(BaseMappingTransformerTransformTests):
    """Tests for the transform method on MappingTransformer."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MappingTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_expected_output(self, library):
        """Test that transform is giving the expected output."""

        df = d.create_df_1(library=library)
        expected = expected_df_1(library=library)

        mapping = {
            "a": {1: "a", 2: "b", 3: "c", 4: "d", 5: "e", 6: "f"},
            "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
        }

        return_dtypes = {"a": "String", "b": "Int8"}

        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

        df = nw.from_native(df)
        expected = nw.from_native(expected)

        # also check single rows
        for i in range(len(df)):
            df_transformed_row = x.transform(df[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_non_specified_values_unchanged(self, library):
        """Test that values not specified in mappings are left unchanged in transform."""

        df = d.create_df_1(library=library)
        expected = expected_df_2(library=library)

        mapping = {"a": {1: 5, 2: 6, 3: 7}, "b": {"a": "z", "b": "y", "c": "x"}}

        return_dtypes = {"a": "Int8", "b": "String"}

        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

        df = nw.from_native(df)
        expected = nw.from_native(expected)

        # also check single rows
        for i in range(len(df)):
            df_transformed_row = x.transform(df[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("mapping", "return_dtypes"),
        [
            ({"a": {1: 1.1, 6: 6.6}}, {"a": "Float64"}),
            (
                {"a": {1: "one", 2: "two", 3: "three", 4: "four", 5: "five", 6: "six"}},
                {"a": "String"},
            ),
            (
                {"a": {1: True, 2: True, 3: True, 4: False, 5: False, 6: False}},
                {"a": "Boolean"},
            ),
            (
                {"b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6}},
                {"b": "Int32"},
            ),
            (
                {"b": {"a": 1.1, "b": 2.2, "c": 3.3, "d": 4.4, "e": 5.5, "f": 6.6}},
                {"b": "Float32"},
            ),
        ],
    )
    def test_expected_dtype_conversions(
        self,
        mapping,
        return_dtypes,
        library,
    ):
        df = d.create_df_1(library=library)
        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)
        df = x.transform(df)

        column = list(mapping.keys())[0]
        actual_dtype = str(nw.from_native(df).get_column(column).dtype)
        assert (
            actual_dtype == return_dtypes[column]
        ), f"dtype converted unexpectedly, expected {return_dtypes[column]} but got {actual_dtype}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_category_dtype_is_conserved(self, library):
        """This is a separate test due to the behaviour of category dtypes.

        See documentation of transform method
        """
        df = d.create_df_1(library=library)
        df = nw.from_native(df)
        df = df.with_columns(nw.col("b").cast(nw.Categorical)).to_native()

        mapping = {"b": {"a": "aaa", "b": "bbb"}}
        return_dtypes = {"b": "Categorical"}

        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)
        df = x.transform(df)

        assert (
            nw.from_native(df).get_column("b").dtype == nw.Categorical
        ), "Categorical dtype not preserved for column b"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("mapping", "mapped_col", "return_dtypes"),
        [
            ({"a": {99: 99, 98: 98}}, "a", {"a": "Int32"}),
            ({"b": {"z": "99", "y": "98"}}, "b", {"b": "String"}),
        ],
    )
    def test_no_applicable_mapping(self, mapping, mapped_col, return_dtypes, library):
        df = d.create_df_1(library=library)

        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)

        with pytest.warns(
            UserWarning,
            match=f"MappingTransformer: No values from mapping for {mapped_col} exist in dataframe.",
        ):
            x.transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("mapping", "mapped_col", "return_dtypes"),
        [
            ({"a": {1: 1, 99: 99}}, "a", {"a": "Int64"}),
            ({"b": {"a": "1", "z": "99"}}, "b", {"b": "String"}),
        ],
    )
    def test_excess_mapping_values(self, mapping, mapped_col, return_dtypes, library):
        df = d.create_df_1(library=library)

        x = MappingTransformer(mappings=mapping, return_dtypes=return_dtypes)

        with pytest.warns(
            UserWarning,
            match=f"MappingTransformer: There are values in the mapping for {mapped_col} that are not present in the dataframe",
        ):
            x.transform(df)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MappingTransformer"



import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    TwoColumnListInitTests,
)
from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import TwoColumnOperatorTransformer


class TestInit(
    NewColumnNameInitMixintests,
    TwoColumnListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "TwoColumnOperatorTransformer"

    def test_axis_not_present_error(self):
        """Checks that an error is raised if no axis element present in pd_method_kwargs dict."""
        with pytest.raises(
            ValueError,
            match='pd_method_kwargs must contain an entry "axis" set to 0 or 1',
        ):
            TwoColumnOperatorTransformer("mul", ["a", "b"], "c", pd_method_kwargs={})

    def test_axis_not_valid_error(self):
        """Checks that an error is raised if no axis element present in pd_method_kwargs dict."""
        with pytest.raises(ValueError, match="pd_method_kwargs 'axis' must be 0 or 1"):
            TwoColumnOperatorTransformer(
                "mul",
                ["a", "b"],
                "c",
                pd_method_kwargs={"axis": 2},
            )


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "TwoColumnOperatorTransformer"

    @pytest.mark.parametrize(
        ("pd_method_name", "output"),
        [
            (
                "mul",
                [4, 10, 18],
            ),
            ("div", [0.25, 0.4, 0.5]),
            ("pow", [1, 32, 729]),
        ],
    )
    def test_expected_output(self, pd_method_name, output):
        """Tests that the output given by TwoColumnOperatorTransformer is as you would expect."""
        expected = d.create_df_11()
        expected["c"] = output
        x = TwoColumnOperatorTransformer(
            pd_method_name,
            ["a", "b"],
            "c",
        )
        actual = x.transform(d.create_df_11())
        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="TwoColumnMethod transformer does not produce the expected output",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "TwoColumnOperatorTransformer"



import re

import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerInitTests,
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import LogTransformer


class TestInit(BaseNumericTransformerInitTests):
    """Tests for LogTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "LogTransformer"

    def test_base_type_error(self):
        """Test that an exception is raised if base is non-numeric."""
        with pytest.raises(
            ValueError,
            match=re.escape("LogTransformer: base should be numeric or None"),
        ):
            LogTransformer(
                columns=["a"],
                base="a",
            )

    def test_suffix_type_error(self):
        """Test that an exception is raised if suffix is non-str."""
        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: suffix should be str",
        ):
            LogTransformer(
                columns=["a"],
                suffix=1,
            )

    def test_add_1_type_error(self):
        """Test that an exception is raised if add_1 is not bool."""
        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: add_1 should be bool",
        ):
            LogTransformer(
                columns=["a"],
                add_1="bla",
            )

    def test_base_not_strictly_positive_error(self):
        """Test that an exception is raised if base is not strictly positive."""
        with pytest.raises(
            ValueError,
            match=re.escape("LogTransformer: base should be strictly positive"),
        ):
            LogTransformer(
                columns=["a"],
                base=0,
            )


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for LogTransformer.transform()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "LogTransformer"

    def expected_df_1():
        """Expected output of test_expected_output_1."""
        df = d.create_df_3()

        df["a_new_col"] = np.log(df["a"])
        df["b_new_col"] = np.log(df["b"])

        return df.drop(columns=["a", "b"])

    def expected_df_2():
        """Expected output of test_expected_output_2."""
        df = d.create_df_3()

        df["a_new_col"] = np.log1p(df["a"])
        df["b_new_col"] = np.log1p(df["b"])

        return df.drop(columns=["a", "b"])

    def expected_df_3():
        """Expected output of test_expected_output_3."""
        df = d.create_df_3()

        df["a_new_col"] = np.log(df["a"])
        df["b_new_col"] = np.log(df["b"])

        return df

    def expected_df_4():
        """Expected output of test_expected_output_4."""
        df = d.create_df_3()

        df["a_new_col"] = np.log1p(df["a"])
        df["b_new_col"] = np.log1p(df["b"])

        return df

    def expected_df_5():
        """Expected output of test_expected_output_5."""
        df = d.create_df_4()

        df["a_new_col"] = np.log(df["a"] + 1) / np.log(5)

        return df

    def expected_df_6():
        """Expected output of test_expected_output_6."""
        df = d.create_df_4()

        df["a_new_col"] = np.log(df["a"]) / np.log(7)

        return df.drop("a", axis=1)

    def test_log1p(self):
        """Test that log1p is working as intended."""
        df = pd.DataFrame(
            {
                "a": [0.00001, 0.00002, 0.00003],
                "b": [0.00004, 0.00005, 0.00006],
            },
        )
        # Values created using np.log1p() of original df
        expected = pd.DataFrame(
            {
                "a_log": [9.999950e-06, 1.999980e-05, 2.999955e-05],
                "b_log": [3.99992000e-05, 4.99987500e-05, 5.99982001e-05],
            },
        )
        log_transformer = LogTransformer(
            columns=["a", "b"],
            add_1=True,
        )
        actual = log_transformer.transform(df)
        pd.testing.assert_frame_equal(actual, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_1()),
    )
    def test_expected_output_1(self, df, expected):
        """Test that transform is giving the expected output when not adding one and dropping original columns."""
        x1 = LogTransformer(
            columns=["a", "b"],
            add_1=False,
            drop_original=True,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform not adding 1 and dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_2()),
    )
    def test_expected_output_2(self, df, expected):
        """Test that transform is giving the expected output when adding one and dropping original columns."""
        x1 = LogTransformer(
            columns=["a", "b"],
            add_1=True,
            drop_original=True,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform adding 1 and dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_3()),
    )
    def test_expected_output_3(self, df, expected):
        """Test that transform is giving the expected output when not adding one and not dropping original columns."""
        x1 = LogTransformer(
            columns=["a", "b"],
            add_1=False,
            drop_original=False,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform not adding 1 and dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_4()),
    )
    def test_expected_output_4(self, df, expected):
        """Test that transform is giving the expected output when adding one and not dropping original columns."""
        x1 = LogTransformer(
            columns=["a", "b"],
            add_1=True,
            drop_original=False,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform not adding 1 and dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_4(), expected_df_5()),
    )
    def test_expected_output_5(self, df, expected):
        """Test that transform is giving the expected output when adding one and not dropping
        original columns and using base.
        """
        x1 = LogTransformer(
            columns=["a"],
            base=5,
            add_1=True,
            drop_original=False,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform not adding 1 and dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_4(), expected_df_6()),
    )
    def test_expected_output_6(self, df, expected):
        """Test that transform is giving the expected output when  not adding one and dropping
        original columns and using base.
        """
        x1 = LogTransformer(
            columns=["a"],
            base=7,
            add_1=False,
            drop_original=True,
            suffix="new_col",
        )

        df_transformed = x1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="LogTransformer transform should be using base, not adding 1, and not dropping original columns",
        )

    @pytest.mark.parametrize(
        ("df", "columns", "add_1", "extra_exception_text"),
        (
            [pd.DataFrame({"a": [1, 2, 0]}), ["a"], False, ""],
            [pd.DataFrame({"a": [1, 2, 0], "b": [1, 2, 3]}), ["a", "b"], False, ""],
            [pd.DataFrame({"a": [1, 2, -1]}), ["a"], True, r" \(after adding 1\)"],
            [
                pd.DataFrame({"a": [1, 2, -1], "b": [1, 2, 3]}),
                ["a", "b"],
                True,
                r" \(after adding 1\)",
            ],
            [pd.DataFrame({"b": [1, 2, -0.001]}), ["b"], False, ""],
            [
                pd.DataFrame({"b": [1, 2, -0.001], "a": [1, 2, 3]}),
                ["a", "b"],
                False,
                "",
            ],
            [pd.DataFrame({"b": [1, 2, -1.001]}), ["b"], True, r" \(after adding 1\)"],
            [
                pd.DataFrame({"b": [1, 2, -1.001], "a": [1, 2, 3]}),
                ["a", "b"],
                True,
                r" \(after adding 1\)",
            ],
        ),
    )
    def test_negative_values_raise_exception(
        self,
        df,
        columns,
        add_1,
        extra_exception_text,
    ):
        """Test that an exception is raised if negative values are passed in transform."""
        x = LogTransformer(columns=columns, add_1=add_1, drop_original=True)

        with pytest.raises(
            ValueError,
            match=f"LogTransformer: values less than or equal to 0 in columns{extra_exception_text}, make greater than 0 before using transform",
        ):
            x.transform(df)



from tests.base_tests import (
    DropOriginalInitMixinTests,
    GenericFitTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    TwoColumnListInitTests,
)
from tests.dates.test_BaseGenericDateTransformer import GenericDatesMixinTransformTests


class TestInit(
    NewColumnNameInitMixintests,
    DropOriginalInitMixinTests,
    TwoColumnListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDateTwoColumnTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDateTwoColumnTransformer"


class TestTransform(GenericTransformTests, GenericDatesMixinTransformTests):
    """Tests for BaseTwoColumnDateTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDateTwoColumnTransformer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDateTwoColumnTransformer"



import pandas as pd
import pytest

from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerFitTests,
    BaseNumericTransformerInitTests,
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import ScalingTransformer


class TestInit(BaseNumericTransformerInitTests):
    """Tests for ScalingTransformer.__init__()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ScalingTransformer"
        cls.scaler_type = "min_max"

    def test_invalid_scaler_type(self):
        """Test that an exception is raised for an invalid scaler type."""
        with pytest.raises(
            ValueError,
            match=r"ScalingTransformer: scaler_type should be one of; \['min_max', 'max_abs', 'standard'\]",
        ):
            ScalingTransformer(columns=["a"], scaler_type="invalid_scaler")


class TestFit(BaseNumericTransformerFitTests):
    """Tests for ScalingTransformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ScalingTransformer"
        cls.scaler_type = "standard"


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for ScalingTransformer.transform()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ScalingTransformer"
        cls.scaler_type = "min_max"

    def test_min_max_scaling(self):
        """Test min-max scaling works correctly."""
        df = pd.DataFrame({"a": [0, 1, 2], "b": [3, 4, 5]})
        transformer = ScalingTransformer(columns=["a", "b"], scaler_type="min_max")
        transformer.fit(df)  # transformer is fitted before transform
        transformed_df = transformer.transform(df)

        expected_df = pd.DataFrame({"a": [0, 0.5, 1], "b": [0, 0.5, 1]})
        pd.testing.assert_frame_equal(transformed_df, expected_df)

    def test_standard_scaling(self):
        """Test standard scaling works correctly."""
        df = pd.DataFrame({"a": [0, 1, 2], "b": [3, 4, 5]})
        transformer = ScalingTransformer(columns=["a", "b"], scaler_type="standard")
        transformer.fit(df)
        transformed_df = transformer.transform(df)

        expected_df = pd.DataFrame(
            {
                "a": [-1.22474487, 0, 1.22474487],  # Standardized values
                "b": [-1.22474487, 0, 1.22474487],
            },
        )
        pd.testing.assert_frame_equal(transformed_df, expected_df)

    def test_max_abs_scaling(self):
        """Test max absolute scaling works correctly."""
        df = pd.DataFrame({"a": [-3, -2, -1], "b": [1, 2, 3]})
        transformer = ScalingTransformer(columns=["a", "b"], scaler_type="max_abs")
        transformer.fit(df)
        transformed_df = transformer.transform(df)

        expected_df = pd.DataFrame({"a": [-1, -2 / 3, -1 / 3], "b": [1 / 3, 2 / 3, 1]})
        pd.testing.assert_frame_equal(transformed_df, expected_df)

    def test_transform_min_max_raises(self):
        """Test that transform scales negative values correctly with MinMaxScaler."""
        df = pd.DataFrame({"a": [-3, -2, -1]})  # Example for Min-Max scalingcode
        transformer = ScalingTransformer(columns=["a"], scaler_type="min_max")
        transformer.fit(df)

        transformed_df = transformer.transform(df)

        # Since MinMaxScaler scales the data to [0, 1] based on the min and max of the input,
        # we calculate the expected scaled values for column 'a'
        expected_df = pd.DataFrame({"a": [0, 0.5, 1]})

        pd.testing.assert_frame_equal(transformed_df, expected_df)

    def test_custom_scaler_args(self):
        """Test that custom arguments passed to the scaler are correctly applied."""
        df = pd.DataFrame({"a": [0, 1, 2]})
        transformer = ScalingTransformer(
            columns=["a"],
            scaler_type="min_max",
            scaler_kwargs={"feature_range": (0, 2)},
        )
        transformer.fit(df)
        transformed_df = transformer.transform(df)

        expected_df = pd.DataFrame({"a": [0, 1, 2]}, dtype=float)
        pd.testing.assert_frame_equal(transformed_df, expected_df, check_dtype=True)



import re

import numpy as np
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
)
from tests.dates.test_BaseDatetimeTransformer import (
    DatetimeMixinTransformTests,
)
from tubular.dates import SeriesDtMethodTransformer


class TestInit(
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    NewColumnNameInitMixintests,
):
    """Tests for SeriesDtMethodTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesDtMethodTransformer"

    def test_invalid_input_type_errors(self):
        """Test that an exceptions are raised for invalid input types."""
        bad_columns = ["b", "c"]
        with pytest.raises(
            ValueError,
            match=rf"SeriesDtMethodTransformer: column should be a str or list of len 1, got {re.escape(str(bad_columns))}",
        ):
            SeriesDtMethodTransformer(
                new_column_name="a",
                pd_method_name=1,
                columns=bad_columns,
            )

        with pytest.raises(
            TypeError,
            match=r"SeriesDtMethodTransformer: unexpected type \(\<class 'int'\>\) for pd_method_name, expecting str",
        ):
            SeriesDtMethodTransformer(
                new_column_name="a",
                pd_method_name=1,
                columns="b",
            )

        with pytest.raises(
            TypeError,
            match=r"""SeriesDtMethodTransformer: pd_method_kwargs should be a dict but got type \<class 'int'\>""",
        ):
            SeriesDtMethodTransformer(
                new_column_name="a",
                pd_method_name="year",
                columns="b",
                pd_method_kwargs=1,
            )

        with pytest.raises(
            TypeError,
            match=r"""SeriesDtMethodTransformer: unexpected type \(\<class 'int'\>\) for pd_method_kwargs key in position 1, must be str""",
        ):
            SeriesDtMethodTransformer(
                new_column_name="a",
                pd_method_name="year",
                columns="b",
                pd_method_kwargs={"a": 1, 2: "b"},
            )

    def test_exception_raised_non_pandas_method_passed(self):
        """Test and exception is raised if a non pd.Series.dt method is passed for pd_method_name."""
        with pytest.raises(
            AttributeError,
            match="""SeriesDtMethodTransformer: error accessing "dt.b" method on pd.Series object - pd_method_name should be a pd.Series.dt method""",
        ):
            SeriesDtMethodTransformer(
                new_column_name="a",
                pd_method_name="b",
                columns="b",
            )


class TestTransform(
    DatetimeMixinTransformTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
):
    """Tests for SeriesDtMethodTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesDtMethodTransformer"

    def expected_df_1():
        """Expected output of test_expected_output_no_overwrite."""
        df = d.create_datediff_test_df()

        df["a_year"] = np.array(
            [1993, 2000, 2018, 2018, 2018, 2018, 2018, 1985],
            dtype=np.int32,
        )

        return df

    def expected_df_2():
        """Expected output of test_expected_output_overwrite."""
        df = d.create_datediff_test_df()

        df["a"] = np.array(
            [1993, 2000, 2018, 2018, 2018, 2018, 2018, 1985],
            dtype=np.int32,
        )

        return df

    def expected_df_3():
        """Expected output of test_expected_output_callable."""
        df = d.create_datediff_test_df()

        df["b_new"] = df["b"].dt.to_period("M")

        return df

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(
            d.create_datediff_test_df(),
            expected_df_1(),
        ),
    )
    def test_expected_output_no_overwrite(self, df, expected):
        """Test a single column output from transform gives expected results, when not overwriting the original column."""
        x = SeriesDtMethodTransformer(
            new_column_name="a_year",
            pd_method_name="year",
            columns="a",
            pd_method_kwargs=None,
        )

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in SeriesDtMethodTransformer.transform with find, not overwriting original column",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(
            d.create_datediff_test_df(),
            expected_df_2(),
        ),
    )
    def test_expected_output_overwrite(self, df, expected):
        """Test a single column output from transform gives expected results, when overwriting the original column."""
        x = SeriesDtMethodTransformer(
            new_column_name="a",
            pd_method_name="year",
            columns="a",
            pd_method_kwargs=None,
        )

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in SeriesDtMethodTransformer.transform with pad, overwriting original column",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(
            d.create_datediff_test_df(),
            expected_df_3(),
        ),
    )
    def test_expected_output_callable(self, df, expected):
        """Test transform gives expected results, when pd_method_name is a callable."""
        x = SeriesDtMethodTransformer(
            new_column_name="b_new",
            pd_method_name="to_period",
            columns="b",
            pd_method_kwargs={"freq": "M"},
        )

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in SeriesDtMethodTransformer.transform with to_period",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesDtMethodTransformer"



import datetime

import narwhals as nw
import pandas as pd
import polars as pl
import pytest

import tests.test_data as d
from tests.base_tests import (
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    TwoColumnListInitTests,
)
from tests.dates.test_BaseGenericDateTransformer import (
    GenericDatesMixinTransformTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.dates import DateDiffLeapYearTransformer


class TestInit(
    NewColumnNameInitMixintests,
    DropOriginalInitMixinTests,
    TwoColumnListInitTests,
):
    """Tests for DateDiffLeapYearTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDiffLeapYearTransformer"

    def test_missing_replacement_type_error(self):
        """Test that an exception is raised if missing_replacement is not the correct type."""
        with pytest.raises(
            TypeError,
            match="DateDiffLeapYearTransformer: if not None, missing_replacement should be an int, float or string",
        ):
            DateDiffLeapYearTransformer(
                columns=["dummy_1", "dummy_2"],
                new_column_name="dummy_3",
                drop_original=False,
                missing_replacement=[1, 2, 3],
            )


def expected_df_1(library="pandas"):
    """Expected output for test_expected_output_drop_original_true."""

    df_dict = {
        "c": [
            26,
            19,
            0,
            0,
            0,
            -2,
            -3,
            30,
        ],
    }

    if library == "pandas":
        return pd.DataFrame(df_dict, dtype="int64[pyarrow]")

    return dataframe_init_dispatch(dataframe_dict=df_dict, library=library)


def expected_df_2(library="pandas"):
    """Expected output for test_expected_output_drop_original_false."""

    df_dict = {
        "a": [
            datetime.date(1993, 9, 27),  # day/month greater than
            datetime.date(2000, 3, 19),  # day/month less than
            datetime.date(2018, 11, 10),  # same day
            datetime.date(2018, 10, 10),  # same year day/month greater than
            datetime.date(2018, 10, 10),  # same year day/month less than
            datetime.date(2018, 10, 10),  # negative day/month less than
            datetime.date(2018, 12, 10),  # negative day/month greater than
            datetime.date(
                1985,
                7,
                23,
            ),  # large gap, this is incorrect with timedelta64 solutions
        ],
        "b": [
            datetime.date(2020, 5, 1),
            datetime.date(2019, 12, 25),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 11, 10),
            datetime.date(2018, 9, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 11, 10),
            datetime.date(2015, 7, 23),
        ],
        "c": [
            26,
            19,
            0,
            0,
            0,
            -2,
            -3,
            30,
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    # ensure types line up with test data
    df = nw.from_native(df)
    for col in [col for col in df.columns if col not in ["c"]]:
        df = df.with_columns(
            nw.col(col).cast(nw.Date),
        )

    if library == "pandas":
        df = nw.to_native(df)
        df["c"] = df["c"].astype("int64[pyarrow]")
        return df

    return nw.to_native(df)


# add the expected to fix float to int with results
def expected_date_diff_df_2(library="pandas"):
    """Expected output for test_expected_output_nans_in_data."""

    df_dict = {
        "c": [
            pd.NA if library == "pandas" else None,
            19,
            0,
            0,
            0,
            -2,
            -3,
            30,
        ],
    }

    if library == "pandas":
        return pd.DataFrame(df_dict, dtype="int64[pyarrow]")

    return pl.DataFrame(df_dict)


# add the expected to fix float to int with results
def expected_date_diff_df_3(library="pandas"):
    """Expected output for test_expected_output_nans_in_data with missing replace with 0."""

    df_dict = {
        "c": [
            0,
            19,
            0,
            0,
            0,
            -2,
            -3,
            30,
        ],
    }

    if library == "pandas":
        return pd.DataFrame(df_dict, dtype="int64[pyarrow]")

    return pl.DataFrame(df_dict)


class TestTransform(
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    GenericDatesMixinTransformTests,
):
    """Tests for DateDiffLeapYearTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDiffLeapYearTransformer"

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (d.create_date_test_df(library="pandas"), expected_df_1(library="pandas")),
            (d.create_date_test_df(library="polars"), expected_df_1(library="polars")),
        ],
    )
    def test_expected_output_drop_original_true(self, df, expected):
        """Test that the output is expected from transform, when drop_original is True.

        This tests positive year gaps, negative year gaps, and missing values.

        """
        x = DateDiffLeapYearTransformer(
            columns=["a", "b"],
            new_column_name="c",
            drop_original=True,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (d.create_date_test_df(library="pandas"), expected_df_2(library="pandas")),
            (d.create_date_test_df(library="polars"), expected_df_2(library="polars")),
        ],
    )
    def test_expected_output_drop_original_false(self, df, expected):
        """Test that the output is expected from transform, when drop_original is False.

        This tests positive year gaps , negative year gaps, and missing values.

        """
        x = DateDiffLeapYearTransformer(
            columns=["a", "b"],
            new_column_name="c",
            drop_original=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("columns"),
        [
            ["date_col_1", "date_col_2"],
            ["datetime_col_1", "datetime_col_2"],
        ],
    )
    def test_expected_output_nans_in_data(self, columns, library):
        "Test that transform works for different date datatype combinations with nans in data"
        x = DateDiffLeapYearTransformer(
            columns=columns,
            new_column_name="c",
            drop_original=True,
        )

        expected = expected_date_diff_df_2(library=library)

        df = d.create_date_diff_different_dtypes_and_nans(library=library)

        df_transformed = x.transform(df[columns])

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_expected_output_nans_in_data_with_replace(self, library):
        "Test that transform works for different date datatype combinations with nans in data and replace nans"
        x = DateDiffLeapYearTransformer(
            columns=["date_col_1", "date_col_2"],
            new_column_name="c",
            drop_original=True,
            missing_replacement=0,
        )

        expected = expected_date_diff_df_3(library=library)

        df = d.create_date_diff_different_dtypes_and_nans(library=library)

        df_transformed = x.transform(df[["date_col_1", "date_col_2"]])

        assert_frame_equal_dispatch(df_transformed, expected)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDiffLeapYearTransformer"



import narwhals as nw
import numpy as np
import pytest

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
    WeightColumnFitMixinTests,
    WeightColumnInitMixinTests,
)
from tests.imputers.test_BaseImputer import (
    GenericImputerTransformTests,
    GenericImputerTransformTestsWeight,
)
from tests.utils import dataframe_init_dispatch
from tubular.imputers import ModeImputer


def input_df_nan(library="pandas"):
    return dataframe_init_dispatch(
        dataframe_dict={"a": [None, None, None]},
        library=library,
    )


class TestInit(ColumnStrListInitTests, WeightColumnInitMixinTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ModeImputer"


class TestFit(WeightColumnFitMixinTests, GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ModeImputer"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_learnt_values(self, library):
        """Test that the impute values learnt during fit are expected."""

        df_dict = {
            # skipping int, as non-nullable for pandas
            # float
            "float_col": [1.0, 2.0, 1.0, None, 3.0],
            # str
            "str_col": ["a", "b", "b", None, None],
            # bool
            "bool_col": [True, True, None, False, True],
            # cat
            "cat_col": ["b", "b", "d", "e", None],
        }

        df = dataframe_init_dispatch(df_dict, library=library)

        # create categorical col
        df = nw.from_native(df)
        df = nw.to_native(df.with_columns(nw.col("cat_col").cast(nw.Categorical)))

        x = ModeImputer(columns=["float_col", "str_col", "cat_col", "bool_col"])

        x.fit(df)

        expected_impute_values = {
            "float_col": 1.0,
            "str_col": "b",
            "cat_col": "b",
            "bool_col": True,
        }

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_learnt_values_tied(self, library):
        """Test that the impute values learnt during fit are expected - when mode is tied."""

        df_dict = {
            # skipping int, as non-nullable for pandas
            # float
            "float_col": [1.0, 2.0, 1.0, None, 2.0],
            # str
            "str_col": ["a", "b", "b", "a", None],
            # bool
            "bool_col": [True, True, None, False, False],
            # cat
            "cat_col": ["b", "b", "e", "e", None],
        }

        df = dataframe_init_dispatch(df_dict, library=library)

        # create categorical col
        df = nw.from_native(df)
        df = nw.to_native(df.with_columns(nw.col("cat_col").cast(nw.Categorical)))

        columns = ["float_col", "str_col", "cat_col", "bool_col"]
        x = ModeImputer(columns=columns)

        x.fit(df)

        expected_impute_values = {
            "float_col": 2.0,
            "str_col": "b",
            "cat_col": "e",
            "bool_col": True,
        }

        with pytest.warns(
            UserWarning,
        ) as warnings:
            x.fit(df)

        for col, w in zip(columns, warnings):
            assert (
                w.message.args[0]
                == f"ModeImputer: The Mode of column {col} is tied, will sort in descending order and return first candidate"
            )

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    @pytest.mark.parametrize(
        ("input_col", "weight_col", "learnt_value", "categorical"),
        [
            # float
            ([1.0, 2.0, None], [2, 2, 2], 2.0, False),
            # str
            (["a", "b", "a", "b", None], [2, 1, 1, 2, 4], "b", False),
            # bool
            ([True, False, None], [4, 4, 1], True, False),
            # cat
            (["g", "g", "h", None], [2, 2, 4, 1], "h", True),
        ],
    )
    def test_learnt_values_tied_weighted(
        self,
        library,
        input_col,
        weight_col,
        learnt_value,
        categorical,
    ):
        """
        Test that the impute values learnt during fit are expected -
        when mode is tied and transformer is weighted.
        """
        df_dict = {
            "col": input_col,
            "weight": weight_col,
        }

        df = dataframe_init_dispatch(df_dict, library=library)

        if categorical:
            # create categorical col
            df = nw.from_native(df)
            df = nw.to_native(df.with_columns(nw.col("col").cast(nw.Categorical)))

        columns = ["col"]
        x = ModeImputer(columns=columns, weights_column="weight")

        x.fit(df)

        expected_impute_values = {
            "col": learnt_value,
        }

        with pytest.warns(
            UserWarning,
        ) as warnings:
            x.fit(df)

        for col, w in zip(columns, warnings):
            assert (
                w.message.args[0]
                == f"ModeImputer: The Mode of column {col} is tied, will sort in descending order and return first candidate"
            )

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_nan_learnt_values(self, library):
        """Test behaviour when learnt value is None."""
        x = ModeImputer(columns=["a"])

        df = input_df_nan(library)

        with pytest.warns(
            UserWarning,
            match="ModeImputer: The Mode of column a is None",
        ):
            x.fit(df)

        expected_impute_values = {"a": None}

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not as expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_nan_learnt_values_weighted(self, library):
        """Test behaviour when learnt value is None - when weights are used."""
        weights_column = "weights_column"
        x = ModeImputer(columns=["a"], weights_column=weights_column)

        df = d.create_weighted_imputers_test_df(library=library)

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        # replace 'a' with all null values to trigger warning
        df = df.with_columns(
            nw.new_series(
                name="a",
                values=[None] * len(df),
                backend=native_backend,
            ),
        )

        df = df.to_native()

        with pytest.warns(
            UserWarning,
            match="ModeImputer: The Mode of column a is None",
        ):
            x.fit(df)

        expected_impute_values = {"a": None}

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not as expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    @pytest.mark.parametrize(
        ("input_col", "weight_col", "learnt_value", "categorical"),
        [
            # float
            ([1.0, 2.0, 2.0, np.nan], [2, 2, 2, 1], 2.0, False),
            # str
            (["a", "b", "a", "b", None, "b"], [2, 1, 1, 2, 4, 3], "b", False),
            # bool
            ([True, False, None, True, True], [4, 4, 1, 1, 1], True, False),
            # cat
            (["a", "b", "c", "c", None], [1, 2, 3, 4, 5], "c", True),
        ],
    )
    def test_learnt_values_weighted_df(
        self,
        library,
        input_col,
        weight_col,
        learnt_value,
        categorical,
    ):
        """Test that the impute values learnt during fit are expected when df is weighted."""
        df_dict = {
            "col": input_col,
            "weight": weight_col,
        }

        df = dataframe_init_dispatch(df_dict, library=library)

        if categorical:
            # create categorical col
            df = nw.from_native(df)
            df = nw.to_native(df.with_columns(nw.col("col").cast(nw.Categorical)))

        columns = ["col"]
        x = ModeImputer(columns=columns, weights_column="weight")

        x.fit(df)

        expected_impute_values = {
            "col": learnt_value,
        }

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_ attribute not as expected, expected {expected_impute_values} but got {x.impute_values_}"


class TestTransform(
    GenericTransformTests,
    GenericImputerTransformTestsWeight,
    GenericImputerTransformTests,
):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ModeImputer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ModeImputer"



from __future__ import annotations

import numpy as np
import pandas as pd


def get_stats_dict(profile_results: list) -> list[dict[str, float]]:
    """extract dict of results from CProfile

    Args:
        profile_results: List of CProfile results

    Returns:
        Dict containing profiling info
    """
    stats_dicts = []
    for profile_result in profile_results:
        stats_dict = {}
        for key, item in profile_result.stats.items():
            new_key = f"{key[0]}.{key[1]}.{key[2]}"
            primitive_calls = item[0]
            calls = item[1]
            total_time = item[2]
            cumulative_time = item[3]
            stats_dict[new_key] = {
                "calls": calls,
                "calls_excl_recursion": primitive_calls,
                "time_excl_subfunctions": total_time,
                "time": cumulative_time,
                "time_per_call": total_time / calls,
            }
        stats_dicts.append(stats_dict)
    stats = [
        "calls",
        "calls_excl_recursion",
        "time_excl_subfunctions",
        "time",
        "time_per_call",
    ]

    # think the profiler will not return below a certain threshold, so the keys vary randomly,
    # find the shared set
    all_keys = [set(stats_dict.keys()) for stats_dict in stats_dicts]
    shared_keys = set.intersection(*all_keys)

    # return stats dict of avg across trials
    return {
        key: {
            stat: np.mean([stats_dict[key][stat] for stats_dict in stats_dicts])
            for stat in stats
        }
        for key in shared_keys
    }


def display_results(stats_dict: dict[str, float], stat: str) -> pd.Series:
    """produce sorted series of runtime info for given runtime statistic

    Args:
        stats_dict: dict of subfunction runtime info
        stat: runtime stat to look at, e.g. 'time'

    Returns:
        pd.Series: sorted series containing requested subfunction runtime info
    """
    stat_dict = {key: stats_dict[key][stat] for key in stats_dict}
    stat_series = pd.Series(stat_dict)
    return stat_series.sort_values(ascending=False)


def add_perc_stats_to_dict(
    stats_dict: dict[str, float],
    overall_time: float,
) -> dict[str, float]:
    """add info on percentage runtime to profiling dict

    Args:
        stats_dict: dict of stats on subfunction runtime info
        overall_time: overall runtime

    Returns:
        Dict on subfunction runtime info, with added percentage runtime info
    """
    for key in stats_dict:
        stats_dict[key]["percentage_time"] = (
            stats_dict[key]["time"] / overall_time
        ) * 100
        stats_dict[key]["percentage_time_per_call"] = (
            stats_dict[key]["percentage_time"] / stats_dict[key]["calls"]
        )
    return stats_dict


def profiling_wrapper(profile_results: list, display_rows: int = 10) -> None:
    """wrapper function to display key profiling info

    Args:
        profile_results: CProfile results
        display_rows: Top n slowest subfunctions to display. Defaults to 10.
    """

    stats_dict = get_stats_dict(profile_results)

    cumulative_time_series = display_results(stats_dict, "time")

    overall_time = cumulative_time_series.iloc[0]
    print("Overall Runtime: ", overall_time)

    stats_dict = add_perc_stats_to_dict(stats_dict, overall_time)

    perc_time_series = display_results(stats_dict, "percentage_time")
    print("Percentage Runtime: \n")
    print(perc_time_series.sort_values(ascending=False).head(display_rows))

    perc_time_per_call_series = display_results(stats_dict, "percentage_time_per_call")
    print("Percentage Runtime Per Call: \n")
    print(perc_time_per_call_series.sort_values(ascending=False).head(display_rows))



import cProfile
import pstats
import time

# stop console filling up with future warnings & mappingtransformer warnings
import warnings

import create_dataset as create
import numpy as np
import profile_helpers as ph
from pipeline_generator import TubularPipelineGenerator

warnings.simplefilter(action="ignore")


if __name__ == "__main__":
    print("***Use time function to inspect transform calls")
    pipe_1 = TubularPipelineGenerator()
    pipe_1 = pipe_1.generate_pipeline()

    df_1 = create.create_standard_pandas_dataset()

    pipe_1.fit(df_1, df_1["AveRooms"])

    pd_rows_to_transform = []
    for _index in range(1000):
        pd_data_row = df_1.iloc[[0]].copy()
        pd_rows_to_transform.append(pd_data_row.copy())

    standard_tubular_times = []

    for i in range(1000):
        start = time.process_time()
        pipe_1.transform(pd_rows_to_transform[i])
        standard_tubular_times.append(time.process_time() - start)

    print("Tubular mean CPU time and SD")
    print(np.mean(standard_tubular_times))
    print(np.std(standard_tubular_times))

    print("***Use cprofile to inspect transform calls")

    df_2 = create.create_standard_pandas_dataset()

    pipe_2 = TubularPipelineGenerator()
    pipe_2 = pipe_2.generate_pipeline()

    pipe_2.fit(df_2, df_2["AveRooms"])

    pd_rows_to_transform_2 = []
    for _index in range(1000):
        pd_data_row = df_2.iloc[[0]].copy()
        pd_rows_to_transform_2.append(pd_data_row.copy())

    cprofile_results = []
    for i in range(1000):
        with cProfile.Profile() as profile:
            pipe_2.transform(pd_rows_to_transform_2[i])

        cprofile_result = pstats.Stats(profile)
        cprofile_results.append(cprofile_result)

    ph.profiling_wrapper(cprofile_results)



import pandas as pd
import pytest

from tubular._utils import (
    _assess_pandas_object_column,  # noqa: PLC2701, importing from private module to test
)


class TestAssessPandasObjectColumn:
    "tests for _assess_pandas_object_column function"

    def test_boolean_case(self):
        "test function output for bool/null column"

        column = "a"

        df = pd.DataFrame({column: [True, False, None]})

        expected_pandas_col_type, expected_polars_col_type = "bool", "Boolean"

        actual_pandas_col_type, actual_polars_col_type = _assess_pandas_object_column(
            df,
            column,
        )

        assert (
            expected_pandas_col_type == actual_pandas_col_type
        ), f"_assess_pandas_object_column output not as expected for pandas bool type, expected {expected_pandas_col_type} but got {actual_pandas_col_type}"

        assert (
            expected_polars_col_type == actual_polars_col_type
        ), f"_assess_pandas_object_column output not as expected for polars bool type, expected {expected_polars_col_type} but got {actual_polars_col_type}"

    def test_null_case(self):
        "test function output for null column"

        column = "a"

        df = pd.DataFrame({column: [None, None]})

        expected_pandas_col_type, expected_polars_col_type = "null", "Unknown"

        actual_pandas_col_type, actual_polars_col_type = _assess_pandas_object_column(
            df,
            column,
        )

        assert (
            expected_pandas_col_type == actual_pandas_col_type
        ), f"_assess_pandas_object_column output not as expected for pandas null type, expected {expected_pandas_col_type} but got {actual_pandas_col_type}"

        assert (
            expected_polars_col_type == actual_polars_col_type
        ), f"_assess_pandas_object_column output not as expected for polars null type, expected {expected_polars_col_type} but got {actual_polars_col_type}"

    def test_object_case(self):
        "test function output for object column (not one of our more specific subcases)"

        column = "a"

        df = pd.DataFrame({column: [["a"], ["b"]]})

        expected_pandas_col_type, expected_polars_col_type = "object", "Object"

        actual_pandas_col_type, actual_polars_col_type = _assess_pandas_object_column(
            df,
            column,
        )

        assert (
            expected_pandas_col_type == actual_pandas_col_type
        ), f"_assess_pandas_object_column output not as expected for pandas null type, expected {expected_pandas_col_type} but got {actual_pandas_col_type}"

        assert (
            expected_polars_col_type == actual_polars_col_type
        ), f"_assess_pandas_object_column output not as expected for polars null type, expected {expected_polars_col_type} but got {actual_polars_col_type}"

    @pytest.mark.parametrize(
        "values",
        [
            [1, 2],
            [True, False],
        ],
    )
    def test_errors_for_non_object(self, values):
        "test function errors for non object columns"

        column = "a"

        df = pd.DataFrame({column: values})

        msg = "_assess_pandas_object_column only works with object dtype columns"

        with pytest.raises(
            TypeError,
            match=msg,
        ):
            _, _ = _assess_pandas_object_column(df, column)



import pytest

from tests.base_tests import OtherBaseBehaviourTests
from tests.capping.test_BaseCappingTransformer import (
    GenericCappingFitTests,
    GenericCappingInitTests,
    GenericCappingTransformTests,
)


class TestInit(GenericCappingInitTests):
    """Tests for CappingTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CappingTransformer"


class TestFit(GenericCappingFitTests):
    """Tests for CappingTransformer.fit()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CappingTransformer"


class TestTransform(GenericCappingTransformTests):
    """Tests for CappingTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CappingTransformer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CappingTransformer"

    def test_get_params_call_with_capping_values_none(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
    ):
        """Test get_params method when capping_values is None."""
        args = minimal_attribute_dict[self.transformer_name]
        args["capping_values"] = None
        args["quantiles"] = {"a": [0.1, 0.9]}
        transformer = uninitialized_transformers[self.transformer_name](**args)

        # Ensure no AttributeError is raised when calling get_params method
        try:
            transformer.get_params()
        except AttributeError as e:
            pytest.fail(f"AttributeError was raised: {e}")



"""This module contains transformers that apply different types of mappings to columns."""

from __future__ import annotations

import warnings
from collections import OrderedDict
from typing import TYPE_CHECKING, Literal, Optional, Union

import narwhals as nw
import numpy as np
import pandas as pd
import polars as pl
from beartype import beartype

from tubular.base import BaseTransformer

if TYPE_CHECKING:
    from narwhals.typing import FrameT


class BaseMappingTransformer(BaseTransformer):
    """Base Transformer Extension for mapping transformers.

    Parameters
    ----------
    mappings : dict
        Dictionary containing column mappings. Each value in mappings should be a dictionary
        of key (column to apply mapping to) value (mapping dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 4}, 'b': {'a': 1, 'b': 2}} would specify
        a mapping for column a of 1->2, 3->4 and a mapping for column b of 'a'->1, b->2.

    return_dtype: Optional[Dict[str, RETURN_DTYPES]]
        Dictionary of col:dtype for returned columns

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    mappings : dict
        Dictionary of mappings for each column individually. The dict passed to mappings in
        init is set to the mappings attribute.

    return_dtypes: dict[str, RETURN_DTYPES]
        Dictionary of col:dtype for returned columns

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    RETURN_DTYPES = Literal[
        "String",
        "Object",
        "Categorical",
        "Boolean",
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "Float32",
        "Float64",
    ]

    @beartype
    def __init__(
        self,
        mappings: dict[str, dict[Union[str, float, int], Union[str, float, int]]],
        return_dtypes: Union[dict[str, RETURN_DTYPES], None] = None,
        **kwargs: Optional[bool],
    ) -> None:
        if not len(mappings) > 0:
            msg = f"{self.classname()}: mappings has no values"
            raise ValueError(msg)

        self.mappings = mappings

        columns = list(mappings.keys())

        # if return_dtypes is not provided, then infer from mappings
        if not return_dtypes:
            return_dtypes = self._infer_return_types(mappings)

        self.return_dtypes = return_dtypes

        super().__init__(columns=columns, **kwargs)

    @staticmethod
    def _infer_return_types(
        mappings: dict[str, dict[str, str | float | int]],
    ) -> dict[str, str]:
        "infer return_dtypes from provided mappings"

        return {col: str(pl.Series(mappings[col].values()).dtype) for col in mappings}

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Base mapping transformer transform method.  Checks that the mappings
        dict has been fitted and calls the BaseTransformer transform method.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd/pl.DataFrame
            Input X, copied if specified by user.

        """
        self.check_is_fitted(["mappings", "return_dtypes"])

        return super().transform(X)


class BaseMappingTransformMixin(BaseTransformer):
    """Mixin class to apply mappings to columns method.

    Transformer uses the mappings attribute which should be a dict of dicts/mappings
    for each required column.

    Attributes
    ----------

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Applies the mapping defined in the mappings dict to each column in the columns
        attribute.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd/pl.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """
        self.check_is_fitted(["mappings", "return_dtypes"])

        X = nw.from_native(super().transform(X))
        native_backend = nw.get_native_namespace(X)

        # will do a join further down, which does not preserve index
        # polars does not care about this, but pandas does,
        # so need to handle a bit carefully
        if nw.get_native_namespace(X).__name__ == "pandas":
            index = nw.to_native(X).index

        # pull out column order to preserve
        column_order = X.columns

        for col in self.mappings:
            mappings = self.mappings[col]

            # TODO - update this logic once narwhals implements map_dict
            # differentiate between unmapped cols and cols mapped to null
            # by including unmapped cols
            unique = X.get_column(col).unique()
            mappings = {key: mappings.get(key, key) for key in unique}

            new_col_values = f"new_{col}_values"
            mappings_df = nw.from_dict(
                {
                    col: list(mappings.keys()),
                    new_col_values: list(mappings.values()),
                },
                schema={
                    col: X.get_column(col).dtype,
                    new_col_values: getattr(nw, self.return_dtypes[col]),
                },
                backend=native_backend,
            )

            X = (
                X.join(
                    mappings_df,
                    how="left",
                    on=col,
                )
                .drop(col)
                .rename({new_col_values: col})
            )

        # restore original index for pandas
        if nw.get_native_namespace(X).__name__ == "pandas":
            X = nw.to_native(X)
            X.index = index

        return X[column_order]


class MappingTransformer(BaseMappingTransformer, BaseMappingTransformMixin):
    """Transformer to map values in columns to other values e.g. to merge two levels into one.

    Note, the MappingTransformer does not require 'self-mappings' to be defined i.e. if you want
    to map a value to itself, you can omit this value from the mappings rather than having to
    map it to itself. This is because it uses the pandas replace method which only replaces values
    which have a corresponding mapping.

    This transformer inherits from BaseMappingTransformMixin as well as the BaseMappingTransformer
    in order to access the pd.Series.replace transform function.

    Parameters
    ----------
    mappings : dict
        Dictionary containing column mappings. Each value in mappings should be a dictionary
        of key (column to apply mapping to) value (mapping dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 4}, 'b': {'a': 1, 'b': 2}} would specify
        a mapping for column a of 1->2, 3->4 and a mapping for column b of 'a'->1, b->2.

    return_dtype: Optional[Dict[str, RETURN_DTYPES]]
        Dictionary of col:dtype for returned columns

    **kwargs
        Arbitrary keyword arguments passed onto BaseMappingTransformer.init method.

    Attributes
    ----------
    mappings : dict
        Dictionary of mappings for each column individually. The dict passed to mappings in
        init is set to the mappings attribute.

    return_dtypes: dict[str, RETURN_DTYPES]
        Dictionary of col:dtype for returned columns

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    @nw.narwhalify
    def transform(
        self,
        X: FrameT,
    ) -> FrameT:
        """Transform the input data X according to the mappings in the mappings attribute dict.

        This method calls the BaseMappingTransformMixin.transform. Note, this transform method is
        different to some of the transform methods in the nominal module, even though they also
        use the BaseMappingTransformMixin.transform method. Here, if a value does not exist in
        the mapping it is unchanged.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd/pl.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """

        BaseTransformer.transform(self, X)

        mapped_columns = self.mappings.keys()

        for col in mapped_columns:
            values_to_be_mapped = set(self.mappings[col].keys())
            values_in_df = set(X.get_column(col).unique())

            if len(values_to_be_mapped.intersection(values_in_df)) == 0:
                warnings.warn(
                    f"{self.classname()}: No values from mapping for {col} exist in dataframe.",
                    stacklevel=2,
                )

            if len(values_to_be_mapped.difference(values_in_df)) > 0:
                warnings.warn(
                    f"{self.classname()}: There are values in the mapping for {col} that are not present in the dataframe",
                    stacklevel=2,
                )

        return nw.from_native(BaseMappingTransformMixin.transform(self, X))


class BaseCrossColumnMappingTransformer(BaseMappingTransformer):
    """BaseMappingTransformer Extension for cross column mapping transformers.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.

    mappings : dict or OrderedDict
        Dictionary containing adjustments. Exact structure will vary by child class.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of mappings for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    def __init__(
        self,
        adjust_column: str,
        mappings: dict[str, dict],
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(mappings=mappings, **kwargs)

        if not isinstance(adjust_column, str):
            msg = f"{self.classname()}: adjust_column should be a string"
            raise TypeError(msg)

        self.adjust_column = adjust_column

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Checks X is valid for transform and calls parent transform

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        X = super().transform(X)

        if self.adjust_column not in X.columns.to_numpy():
            msg = f"{self.classname()}: variable {self.adjust_column} is not in X"
            raise ValueError(msg)

        return X


class CrossColumnMappingTransformer(BaseCrossColumnMappingTransformer):
    """Transformer to adjust values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.

    mappings : dict or OrderedDict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 'a', 3: 'b'}, 'b': {'a': 1, 'b': 2}}
        would replace the values in the adjustment column based off the values in column a using the mapping
        1->'a', 3->'b' and also replace based off the values in column b using a mapping 'a'->1, 'b'->2.
        If more than one column is defined for this mapping, then this object must be an OrderedDict
        to ensure reproducibility.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of mappings for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    def __init__(
        self,
        adjust_column: str,
        mappings: dict[str, dict],
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(mappings=mappings, adjust_column=adjust_column, **kwargs)

        if len(mappings) > 1 and not isinstance(mappings, OrderedDict):
            msg = f"{self.classname()}: mappings should be an ordered dict for 'replace' mappings using multiple columns"
            raise TypeError(msg)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        X = super().transform(X)

        for i in self.columns:
            for j in self.mappings[i]:
                X[self.adjust_column] = np.where(
                    (X[i] == j),
                    self.mappings[i][j],
                    X[self.adjust_column],
                )

        return X


class BaseCrossColumnNumericTransformer(BaseCrossColumnMappingTransformer):
    """BaseCrossColumnNumericTransformer Extension for cross column numerical mapping transformers.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.

    mappings : dict
        Dictionary containing adjustments. Exact structure will vary by child class.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of mappings for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    def __init__(
        self,
        adjust_column: str,
        mappings: dict[str, dict],
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(mappings=mappings, adjust_column=adjust_column, **kwargs)

        for j in mappings.values():
            for k in j.values():
                if type(k) not in [int, float]:
                    msg = f"{self.classname()}: mapping values must be numeric"
                    raise TypeError(msg)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Checks X is valid for transform and calls parent transform

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        X = super().transform(X)

        if not pd.api.types.is_numeric_dtype(X[self.adjust_column]):
            msg = f"{self.classname()}: variable {self.adjust_column} must have numeric dtype."
            raise TypeError(msg)

        return X


class CrossColumnMultiplyTransformer(BaseCrossColumnNumericTransformer):
    """Transformer to apply a multiplicative adjustment to values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.  The data type of this column must be int or float.

    mappings : dict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 5}, 'b': {'a': 0.5, 'b': 1.1}}
        would multiply the values in the adjustment column based off the values in column a using the mapping
        1->2*value, 3->5*value and also multiply based off the values in column b using a mapping
        'a'->0.5*value, 'b'->1.1*value.
        The values within the dicts defining the multipliers must have type int or float.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of multiplicative adjustments for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework


    """

    polars_compatible = False

    def __init__(
        self,
        adjust_column: str,
        mappings: dict[str, dict],
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(mappings=mappings, adjust_column=adjust_column, **kwargs)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        X = super().transform(X)

        for i in self.columns:
            for j in self.mappings[i]:
                X[self.adjust_column] = np.where(
                    (X[i] == j),
                    X[self.adjust_column] * self.mappings[i][j],
                    X[self.adjust_column],
                )

        return X


class CrossColumnAddTransformer(BaseCrossColumnNumericTransformer):
    """Transformer to apply an additive adjustment to values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.  The data type of this column must be int or float.

    mappings : dict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 5}, 'b': {'a': 1, 'b': -5}}
        would provide an additive adjustment to the values in the adjustment column based off the values
        in column a using the mapping 1->2+value, 3->5+value and also an additive adjustment based off the
        values in column b using a mapping 'a'->1+value, 'b'->(-5)+value.
        The values within the dicts defining the values to be added must have type int or float.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of additive adjustments for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework


    """

    polars_compatible = False

    def __init__(
        self,
        adjust_column: str,
        mappings: dict[str, dict],
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(mappings=mappings, adjust_column=adjust_column, **kwargs)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        X = super().transform(X)

        for i in self.columns:
            for j in self.mappings[i]:
                X[self.adjust_column] = np.where(
                    (X[i] == j),
                    X[self.adjust_column] + self.mappings[i][j],
                    X[self.adjust_column],
                )

        return X



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("../../"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "tubular"
project_copyright = "2025, LV GI Data Science Team"
author = "LV GI Data Science Team"
release = "tubular.__version__"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
]


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import datetime

import numpy as np
import pandas as pd
import polars as pl

from tests.utils import dataframe_init_dispatch


def test_output_types():
    "simple sanity check that requested df type is returned"

    df_dict = {"a": [1, 2], "b": [3, 4]}

    pandas_output = dataframe_init_dispatch(library="pandas", dataframe_dict=df_dict)

    assert isinstance(
        pandas_output,
        pd.DataFrame,
    ), "dataframe_init_dispatch not returning pandas df for library=pandas"

    polars_output = dataframe_init_dispatch(library="polars", dataframe_dict=df_dict)

    assert isinstance(
        polars_output,
        pl.DataFrame,
    ), "dataframe_init_dispatch not returning polars df for library=polars"


def test_type_alignment():
    "test that types are matched as expected between polars/pandas requests"

    df_dict = {
        # int
        "a": [1, 2],
        # int with null (float)
        "a1": [1, None],
        # float
        "b": [3.0, 4.0],
        # float with None
        "c": [3.0, np.nan],
        # str
        "d": ["a", "b"],
        # str with None
        "e": ["a", None],
        # bool
        "f": [True, False],
        # bool with None
        "g": [True, None],
        # None
        "h": [None, None],
        # date
        "i": [datetime.date(2020, 5, 1), datetime.date(2021, 5, 1)],
        # datetime
        "j": [
            datetime.datetime(1993, 9, 27, tzinfo=datetime.timezone.utc),
            datetime.datetime(1995, 9, 27, tzinfo=datetime.timezone.utc),
        ],
    }

    expected_pandas_types = {
        "a": np.int64,
        "a1": np.float64,
        "b": np.float64,
        "c": np.float64,
        "d": object,
        "e": object,
        "f": bool,
        "g": object,
        "h": object,
        "i": "date32[day][pyarrow]",
        "j": "datetime64[ns, UTC]",
    }

    expected_polars_types = {
        "a": pl.Int64,
        "a1": pl.Float64,  # function is designed to copy pandas handling
        "b": pl.Float64,
        "c": pl.Float64,
        "d": pl.String,
        "e": pl.String,
        "f": pl.Boolean,
        "g": pl.Boolean,
        "h": pl.Null,
        "i": pl.Date,
        "j": pl.Datetime,
    }

    pandas_output = dataframe_init_dispatch(library="pandas", dataframe_dict=df_dict)

    polars_output = dataframe_init_dispatch(library="polars", dataframe_dict=df_dict)

    assert (
        set(polars_output.columns) == set(pandas_output.columns)
    ), "dataframe_init_dispatch: polars and pandas output should have same columns for same input dataframe_dict"

    for col in pandas_output:
        assert pandas_output[col].dtype == expected_pandas_types[col]

        assert polars_output[col].dtype == expected_polars_types[col]



import re

import numpy as np
import pytest
import test_aide as ta
from beartype.roar import BeartypeCallHintParamViolation

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
)
from tubular.strings import SeriesStrMethodTransformer


class TestInit(ColumnStrListInitTests, NewColumnNameInitMixintests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesStrMethodTransformer"

    # Duplicated here as base test sets "columns" to a list of len(2), not viable for this transformer
    @pytest.mark.parametrize(
        "non_string",
        [1, True, {"a": 1}, [1, 2], None, np.inf, np.nan],
    )
    def test_columns_list_element_error(
        self,
        non_string,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if columns list contains non-string elements."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = [non_string]

        with pytest.raises(
            BeartypeCallHintParamViolation,
        ):
            uninitialized_transformers[self.transformer_name](**args)

    def test_list_length(self):
        """Test that an error is raised if columns list contains more than a single element"""

        with pytest.raises(
            ValueError,
            match="SeriesStrMethodTransformer: columns arg should contain only 1 column name but got 2",
        ):
            SeriesStrMethodTransformer(
                new_column_name="a",
                pd_method_name=1,
                columns=["b", "c"],
            )

    def test_exception_raised_non_pandas_method_passed(self):
        """Test and exception is raised if a non pd.Series.str method is passed for pd_method_name."""
        with pytest.raises(
            AttributeError,
            match="""SeriesStrMethodTransformer: error accessing "str.b" method on pd.Series object - pd_method_name should be a pd.Series.str method""",
        ):
            SeriesStrMethodTransformer(
                new_column_name="a",
                pd_method_name="b",
                columns=["b"],
            )

    @pytest.mark.parametrize(
        "non_dict",
        [1, "a", True, [1, 2], np.inf, np.nan],
    )
    def test_invalid_pd_kwargs_type_errors(
        self,
        non_dict,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exceptions are raised for invalid pd_kwargs types."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["pd_method_kwargs"] = non_dict

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: pd_method_kwargs should be provided as a dict or defaulted to None",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)

    @pytest.mark.parametrize(
        "na_dict_key",
        [{"a": 1, 2: "b"}, {"a": 1, (1, 2): "b"}],
    )
    def test_invalid_pd_kwargs_key_errors(
        self,
        na_dict_key,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exceptions are raised for invalid pd_kwargs key types."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["pd_method_kwargs"] = na_dict_key

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: all keys in pd_method_kwargs must be a string value",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)


class TestTransform(GenericTransformTests):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesStrMethodTransformer"

    def expected_df_1():
        """Expected output of test_expected_output_no_overwrite."""
        df = d.create_df_7()

        df["b_new"] = df["b"].str.find(sub="a")

        return df

    def expected_df_2():
        """Expected output of test_expected_output_overwrite."""
        df = d.create_df_7()

        df["b"] = df["b"].str.pad(width=10)

        return df

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_7(), expected_df_1()),
    )
    def test_expected_output_no_overwrite(self, df, expected):
        """Test a single column output from transform gives expected results, when not overwriting the original column."""
        x = SeriesStrMethodTransformer(
            new_column_name="b_new",
            pd_method_name="find",
            columns=["b"],
            pd_method_kwargs={"sub": "a"},
        )

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in SeriesStrMethodTransformer.transform with find, not overwriting original column",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_7(), expected_df_2()),
    )
    def test_expected_output_overwrite(self, df, expected):
        """Test a single column output from transform gives expected results, when overwriting the original column."""
        x = SeriesStrMethodTransformer(
            new_column_name="b",
            pd_method_name="pad",
            columns=["b"],
            pd_method_kwargs={"width": 10},
        )

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in SeriesStrMethodTransformer.transform with pad, overwriting original column",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SeriesStrMethodTransformer"
